import typing

from ..geometry import Rotation2d


class ChassisSpeeds(typing.NamedTuple):
    """Represents the speed of a robot chassis.

    Although this struct contains similar members compared to a
    Twist2d, they do NOT represent the same thing.  Whereas a Twist2d
    represents a change in pose w.r.t. to the robot frame of reference,
    this ChassisSpeeds struct represents a velocity w.r.t. to the robot
    frame of reference.

    A strictly non-holonomic drivetrain, such as a differential drive, should
    never have a dy component because it can never move sideways. Holonomic
    drivetrains such as swerve and mecanum will often have all three components.
    """

    #: Represents forward velocity w.r.t the robot frame of reference. (Fwd is +)
    vx: float = 0
    #: Represents strafe velocity w.r.t the robot frame of reference. (Left is +)
    vy: float = 0
    #: Represents the angular velocity of the robot frame. (CCW is +)
    omega: float = 0

    @classmethod
    def fromFieldRelativeSpeeds(
        cls, vx: float, vy: float, omega: float, robotAngle: Rotation2d
    ) -> "ChassisSpeeds":
        """Converts a user provided field-relative set of speeds into a robot-relative
        ChassisSpeeds object.

        :param vx: The component of speed in the x direction relative to the field.
                   Positive x is away from your alliance wall.

        :param vy: The component of speed in the y direction relative to the field.
                   Positive y is to your left when standing behind the alliance wall.

        :param omega: The angular rate of the robot.

        :param robotAngle: The angle of the robot as measured by a gyroscope.
            The robot's angle is considered to be zero when it is facing
            directly away from your alliance station wall.
            Remember that this should be CCW positive.

        :returns: Object representing the speeds in the robot's frame of reference.
        """
        return cls(
            vx * robotAngle.cos + vy * robotAngle.sin,
            -vx * robotAngle.sin + vy * robotAngle.cos,
            omega,
        )



import math
from dataclasses import dataclass
from typing import overload

from .rotation2d import Rotation2d
from .translation2d import Translation2d
from .twist2d import Twist2d

__all__ = ("Rotation2d", "Translation2d", "Twist2d", "Transform2d", "Pose2d")

_identity_translation = Translation2d()
_zero_rotation = Rotation2d()


@dataclass
class Transform2d:
    """Represents a transformation for a Pose2d."""

    translation: Translation2d
    rotation: Rotation2d

    __slots__ = ("translation", "rotation")

    def __init__(
        self,
        translation: Translation2d = _identity_translation,
        rotation: Rotation2d = _zero_rotation,
    ):
        self.translation = translation
        self.rotation = rotation

    def __mul__(self, other: float) -> "Transform2d":
        """Multiplies the transform by a scalar."""
        if not isinstance(other, (float, int)):
            return NotImplemented
        return Transform2d(self.translation * other, self.rotation * other)

    __rmul__ = __mul__


@dataclass
class Pose2d:
    """Represents a 2d pose containing translational and rotational elements."""

    translation: Translation2d
    rotation: Rotation2d

    __slots__ = ("translation", "rotation")

    @overload
    def __init__(self):
        """Constructs a pose at the origin facing toward the positive X axis."""

    @overload
    def __init__(self, translation: Translation2d, rotation: Rotation2d):
        """Constructs a pose with the specified translation and rotation."""

    @overload
    def __init__(self, x: float, y: float, rotation: Rotation2d):
        """Convenience constructor that takes x and y values directly
        instead of requiring a Translation2d.
        """

    def __init__(self, *args, rotation=None):
        if rotation is None:
            if args:
                *args, rotation = args
            else:
                rotation = _zero_rotation

        if len(args) == 1:
            translation = args[0]
            assert isinstance(translation, Translation2d)
        else:
            translation = Translation2d(*args)

        self.translation = translation
        self.rotation = rotation

    def __add__(self, other: "Transform2d") -> "Pose2d":
        """Transforms the pose by the given transformation.

        ::

            [x_new]   [cos, -sin, 0][transform.x]
            [y_new] = [sin,  cos, 0][transform.y]
            [t_new]   [0,    0,   1][transform.t]

        :param other: The transform to transform the pose by.

        :returns: The transformed pose.
        """
        if not isinstance(other, Transform2d):
            return NotImplemented
        return Pose2d(
            self.translation + other.translation.rotateBy(self.rotation),
            self.rotation + other.rotation,
        )

    def __sub__(self, other: "Pose2d") -> "Transform2d":
        """Returns the Transform2d that maps the other pose to self.

        :param other: The initial pose of the transformation.

        :returns: The transform that maps the other pose to the current pose.
        """
        if not isinstance(other, Pose2d):
            return NotImplemented
        # We are rotating the difference between the translations
        # using a clockwise rotation matrix. This transforms the global
        # delta into a local delta (relative to the initial pose).
        translation = (self.translation - other.translation).rotateBy(-other.rotation)
        rotation = self.rotation - other.rotation
        return Transform2d(translation, rotation)

    def relativeTo(self, other: "Pose2d") -> "Pose2d":
        """Returns the other pose relative to the current pose.

        This function can often be used for trajectory tracking or pose
        stabilization algorithms to get the error between the reference
        and the current pose.

        :param other: The pose that is the origin of the new coordinate
                      frame that the current pose will be converted into.

        :returns: The current pose relative to the new origin pose.
        """
        transform = self - other
        return Pose2d(transform.translation, transform.rotation)

    def exp(self, twist: Twist2d) -> "Pose2d":
        """Obtain a new Pose2d from a (constant curvature) velocity.

        See <https://file.tavsys.net/control/state-space-guide.pdf>
        section on nonlinear pose estimation for derivation.

        The twist is a change in pose in the robot's coordinate frame
        since the previous pose update. When the user runs exp() on the
        previous known field-relative pose with the argument being the
        twist, the user will receive the new field-relative pose.

        "Exp" represents the pose exponential, which is solving a
        differential equation moving the pose forward in time.

        :param twist: The change in pose in the robot's coordinate frame
             since the previous pose update. For example, if a
             non-holonomic robot moves forward 0.01 meters and changes
             angle by 0.5 degrees since the previous pose update, the
             twist would be ``Twist2d(0.01, 0.0, math.radians(0.5))``

        :returns: The new pose of the robot.
        """
        dx = twist.dx
        dy = twist.dy
        dtheta = twist.dtheta

        sin_theta = math.sin(dtheta)
        cos_theta = math.cos(dtheta)

        if abs(dtheta) < 1e-9:
            s = 1.0 - 1 / 6 * dtheta ** 2
            c = 0.5 * dtheta
        else:
            s = sin_theta / dtheta
            c = (1.0 - cos_theta) / dtheta

        transform = Transform2d(
            Translation2d(dx * s - dy * c, dx * c + dy * s),
            Rotation2d(cos_theta, sin_theta),
        )
        return self + transform

    def log(self, end: "Pose2d") -> Twist2d:
        """Returns a Twist2d that maps this pose to the end pose.

        If c = a.log(b), then a.exp(c) = b.

        :param end: The end pose for the transformation.

        :returns: The twist that maps self to end.
        """
        transform = end - self
        dtheta = transform.rotation.value
        half_dtheta = 0.5 * dtheta

        cos_minus_one = transform.rotation.cos - 1.0

        if abs(cos_minus_one) < 1e-9:
            half_theta_by_tan_half_dtheta = 1.0 - 1 / 12 * dtheta ** 2
        else:
            half_theta_by_tan_half_dtheta = (
                -(half_dtheta * transform.rotation.sin) / cos_minus_one
            )

        translation_part = transform.translation.rotateBy(
            Rotation2d(half_theta_by_tan_half_dtheta, -half_dtheta)
        ) * math.hypot(half_theta_by_tan_half_dtheta, half_dtheta)
        return Twist2d(translation_part.x, translation_part.y, dtheta)



import math
from dataclasses import dataclass
from typing import List, Optional

import numpy as np

from ..geometry import (
    Pose2d,
    Rotation2d,
    Translation2d,
    Twist2d,
    _identity_translation,
    _zero_rotation,
)
from .chassisspeeds import ChassisSpeeds


__all__ = ("SwerveModuleState", "SwerveDriveKinematics", "SwerveDriveOdometry")


@dataclass
class SwerveModuleState:
    """Represents the state of one swerve module."""

    #: Speed of the wheel of the module.
    speed: float
    #: Angle of the module.
    angle: Rotation2d

    __slots__ = ("speed", "angle")

    def __init__(self, speed: float = 0, angle: Rotation2d = _zero_rotation):
        self.speed = speed
        self.angle = angle


class SwerveDriveKinematics:
    """Helper class that converts a chassis velocity (dx, dy, and dtheta components)
    into individual module states (speed and angle).

    The inverse kinematics (converting from a desired chassis velocity to
    individual module states) uses the relative locations of the modules with
    respect to the center of rotation. The center of rotation for inverse
    kinematics is also variable. This means that you can set your set your center
    of rotation in a corner of the robot to perform special evasion maneuvers.

    Forward kinematics (converting an array of module states into the overall
    chassis motion) is performs the exact opposite of what inverse kinematics
    does. Since this is an overdetermined system (more equations than variables),
    we use a least-squares approximation.

    The inverse kinematics: [moduleStates] = [moduleLocations] * [chassisSpeeds]
    We take the Moore-Penrose pseudoinverse of [moduleLocations] and then
    multiply by [moduleStates] to get our chassis speeds.

    Forward kinematics is also used for odometry -- determining the
    position of the robot on the field using encoders and a gyro.
    """

    __slots__ = (
        "modules",
        "num_modules",
        "_inverse_kinematics",
        "forward_kinematics",
        "_prev_cor",
    )

    def __init__(self, *wheels: Translation2d):
        """Constructs a swerve drive kinematics object.

        This takes in a variable number of wheel locations as Translation2ds.
        The order in which you pass in the wheel locations is the same order that
        you will receive the module states when performing inverse kinematics.
        It is also expected that you pass in the module states in the
        same order when calling the forward kinematics methods.

        :param wheels: The locations of the wheels relative to the
                       physical center of the robot.
        """
        assert len(wheels) >= 2, "A swerve drive requires at least two modules"

        inverse_kinematics = np.array([(1, 0, 0), (0, 1, 0)] * len(wheels), dtype=float)
        for i, module in enumerate(wheels):
            inverse_kinematics[2 * i, 2] = -module.y
            inverse_kinematics[2 * i + 1, 2] = module.x

        self.modules = wheels
        self.num_modules = len(wheels)
        self._inverse_kinematics = inverse_kinematics
        self.forward_kinematics = np.linalg.pinv(inverse_kinematics)
        self._prev_cor = _identity_translation

    def toSwerveModuleStates(
        self,
        chassisSpeeds: ChassisSpeeds,
        centerOfRotation: Translation2d = _identity_translation,
    ) -> List[SwerveModuleState]:
        """Performs inverse kinematics to return the module states from a desired
        chassis velocity.

        This method is often used to convert joystick values into
        module speeds and angles.

        This function also supports variable centers of rotation.
        During normal operations, the center of rotation is usually the
        same as the physical center of the robot; therefore, the argument
        is defaulted to that use case.  However, if you wish to change
        the center of rotation for evasive maneuvers, vision alignment,
        or for any other use case, you can do so.

        :param chassisSpeeds: The desired chassis speed.

        :param centerOfRotation: The center of rotation.
            For example, if you set the center of rotation at one corner
            of the robot and provide a chassis speed that only has a
            dtheta component, the robot will rotate around that corner.

        :returns: An array containing the module states.
            Use caution because these module states are not normalized.
            Sometimes, a user input may cause one of the module speeds
            to go above the attainable max velocity.
            Use the :meth:`normalizeWheelSpeeds` function to rectify this issue.
        """
        prev_cor = self._prev_cor
        inverse_kinematics = self._inverse_kinematics
        if prev_cor is not centerOfRotation and prev_cor != centerOfRotation:
            cor_y = centerOfRotation.y
            cor_x = centerOfRotation.x
            for i, module in enumerate(self.modules):
                inverse_kinematics[2 * i, 2] = -module.y + cor_y
                inverse_kinematics[2 * i + 1, 2] = module.x - cor_x
            self._prev_cor = centerOfRotation

        chassis_vel_vec = np.array(chassisSpeeds)
        module_states = inverse_kinematics @ chassis_vel_vec
        return [
            SwerveModuleState(math.hypot(x, y), Rotation2d(x, y))
            for x, y in module_states.reshape(-1, 2)
        ]

    def toChassisSpeeds(self, *wheel_states: SwerveModuleState) -> ChassisSpeeds:
        """Performs forward kinematics to return the resulting chassis state
        from the given module states.

        This method is often used for odometry -- determining the robot's
        position on the field using data from the real-world speed and
        angle of each module on the robot.

        :param wheel_states: The state of the modules (as a SwerveModuleState type)
                             as measured from respective encoders and gyros.
                             The order of the swerve module states should be
                             same as passed into the constructor of this class.

        :returns: The resulting chassis speed.
        """
        num_modules = self.num_modules
        assert (
            len(wheel_states) == num_modules
        ), "Number of modules must be consistent with number of wheel locations."
        module_states_mat = np.array(
            [
                (module.speed * module.angle.cos, module.speed * module.angle.sin)
                for module in wheel_states
            ]
        ).reshape(-1)
        chassis_vel_vec = self.forward_kinematics @ module_states_mat
        return ChassisSpeeds(*chassis_vel_vec)

    @staticmethod
    def normalizeWheelSpeeds(
        moduleStates: List[SwerveModuleState], attainableMaxSpeed: float
    ) -> None:
        """Normalizes the wheel speeds using some max attainable speed.

        Sometimes, after inverse kinematics, the requested speed from
        a/several modules may be above the max attainable speed for
        the driving motor on that module. To fix this issue, one can
        "normalize" all the wheel speeds to make sure that all requested
        module speeds are below the absolute threshold, while maintaining
        the ratio of speeds between modules.

        :param moduleStates: Reference to list of module states.
                             The list will be mutated with the normalized speeds!

        :param attainableMaxSpeed: The absolute max speed that a module can reach.
        """
        real_max_speed = max(module.speed for module in moduleStates)
        if real_max_speed > attainableMaxSpeed:
            factor = attainableMaxSpeed / real_max_speed
            for module in moduleStates:
                # module.speed = module.speed / real_max_speed * attainableMaxSpeed
                module.speed *= factor


class SwerveDriveOdometry:
    """Class for swerve drive odometry.

    Odometry allows you to track the robot's position on the field over
    a course of a match using readings from your swerve drive encoders
    and swerve azimuth encoders.

    Teams can use odometry during the autonomous period for complex
    tasks like path following. Furthermore, odometry can be used for
    latency compensation when using computer-vision systems.
    """

    __slots__ = (
        "kinematics",
        "_pose",
        "_previous_time",
        "_previous_angle",
        "_gyro_offset",
    )

    def __init__(
        self,
        kinematics: SwerveDriveKinematics,
        gyroAngle: Rotation2d,
        initialPose: Optional[Pose2d] = None,
    ):
        if initialPose is None:
            initialPose = Pose2d()

        self.kinematics = kinematics
        self._pose = initialPose
        self._previous_time: Optional[float] = None
        self._previous_angle = initialPose.rotation
        self._gyro_offset = initialPose.rotation - gyroAngle

    def resetPosition(self, pose: Pose2d, gyroAngle: Rotation2d) -> None:
        """Resets the robot's position on the field.

        The gyroscope angle does not need to be reset here on the user's robot
        code. The library automatically takes care of offsetting the gyro angle.

        :param pose: The position on the field that your robot is at.

        :param gyroAngle: The angle reported by the gyroscope.
        """
        self._pose = pose
        self._previous_angle = pose.rotation
        self._gyro_offset = pose.rotation - gyroAngle

    def getPose(self) -> Pose2d:
        """Returns the position of the robot on the field."""
        return self._pose

    def updateWithTime(
        self,
        currentTime: float,
        gyroAngle: Rotation2d,
        *module_states: SwerveModuleState,
    ) -> Pose2d:
        """Updates the robot's position on the field using forward kinematics
        and integration of the pose over time.

        This method takes in the current time as a parameter to calculate
        period (difference between two timestamps). The period is used
        to calculate the change in distance from a velocity. This also
        takes in an angle parameter which is used instead of the angular
        rate that is calculated from forward kinematics.

        :param currentTime: The current time.

        :param gyroAngle: The angle reported by the gyroscope.

        :param module_states: The current state of all swerve modules.
                    Please provide the states in the same order in which
                    you instantiated your SwerveDriveKinematics.

        :returns: The new pose of the robot.
        """
        prev_time = self._previous_time
        delta_time = currentTime - prev_time if prev_time is not None else 0
        self._previous_time = currentTime

        angle = gyroAngle + self._gyro_offset
        dx, dy, _dtheta = self.kinematics.toChassisSpeeds(*module_states)

        new_pose = self._pose.exp(
            Twist2d(
                dx * delta_time,
                dy * delta_time,
                (angle - self._previous_angle).getRadians(),
            )
        )

        self._previous_angle = angle
        self._pose = Pose2d(new_pose.translation, angle)

        return self._pose



import math
from dataclasses import dataclass


@dataclass
class Twist2d:
    """A change in distance along arc since the last pose update.

    We can use ideas from differential calculus to create new Pose2ds
    from a Twist2d and vice versa.

    A Twist can be used to represent a difference between two poses.
    """

    #: Linear "dx" component
    dx: float
    #: Linear "dy" component
    dy: float
    #: Angular "dtheta" component (radians)
    dtheta: float

    __slots__ = ("dx", "dy", "dtheta")

    def __init__(self, dx: float = 0, dy: float = 0, dtheta: float = 0):
        self.dx = dx
        self.dy = dy
        self.dtheta = dtheta

    def __eq__(self, other: "Twist2d") -> bool:
        if not isinstance(other, Twist2d):
            return NotImplemented
        return (
            math.isclose(self.dx, other.dx)
            and math.isclose(self.dy, other.dy)
            and math.isclose(self.dtheta, other.dtheta)
        )



import math
from dataclasses import dataclass
from typing import overload


@dataclass(frozen=True)
class Rotation2d:
    """A rotation in a 2d coordinate frame represented a point on the unit circle."""

    #: The value of the rotation in radians.
    value: float
    #: The cosine of the rotation.
    cos: float
    #: The sine of the rotation.
    sin: float

    __slots__ = ("value", "cos", "sin")

    @overload
    def __init__(self, __value: float = 0):
        """Constructs a Rotation2d with the given radian value."""

    @overload
    def __init__(self, __x: float, __y: float):
        """Constructs a Rotation2d with the given cosine and sine components.

        The x and y don't have to be normalised.
        """

    def __init__(self, *args: float):
        """Constructs a Rotation2d.

        This can either be called with zero or one arguments:

        :param float value: The value in radians (default 0).

        ... or with two arguments:

        :param float x: The x component or cosine of the rotation.
        :param float y: The y component or sine of the rotation.

        In this case the x and y do not need to be normalised.

        """
        value: float
        cos: float
        sin: float

        if not args:
            value = 0
            cos = 1
            sin = 0
        elif len(args) == 1:
            value = args[0]
            cos = math.cos(value)
            sin = math.sin(value)
        else:
            x, y = args
            magnitude = math.hypot(x, y)
            if magnitude > 1e-6:
                sin = y / magnitude
                cos = x / magnitude
            else:
                sin = 0
                cos = 1
            value = math.atan2(sin, cos)

        object.__setattr__(self, "value", value)
        object.__setattr__(self, "cos", cos)
        object.__setattr__(self, "sin", sin)

    @classmethod
    def fromDegrees(cls, degrees: float) -> "Rotation2d":
        """Creates a Rotation2d with the given degrees value."""
        return cls(math.radians(degrees))

    def __repr__(self) -> str:
        return f"Rotation2d({self.value})"

    def __add__(self, other: "Rotation2d") -> "Rotation2d":
        """Adds two rotations together, with the result bounded between -pi and pi."""
        if not isinstance(other, Rotation2d):
            return NotImplemented
        cos_a = self.cos
        cos_b = other.cos
        sin_a = self.sin
        sin_b = other.sin
        return Rotation2d(cos_a * cos_b - sin_a * sin_b, cos_a * sin_b + sin_a * cos_b)

    def __sub__(self, other: "Rotation2d") -> "Rotation2d":
        """Subtracts the other rotation from self."""
        if not isinstance(other, Rotation2d):
            return NotImplemented
        return self + -other

    def __neg__(self) -> "Rotation2d":
        """Takes the inverse of the current rotation.

        This is simply the negative of the current angular value.
        """
        return Rotation2d(-self.value)

    def __mul__(self, other: float) -> "Rotation2d":
        """Multiplies the current rotation by a scalar."""
        if not isinstance(other, (float, int)):
            return NotImplemented
        return Rotation2d(self.value * other)

    __rmul__ = __mul__

    def __eq__(self, other: "Rotation2d") -> bool:
        if not isinstance(other, Rotation2d):
            return NotImplemented
        return math.isclose(self.value, other.value)

    def getRadians(self) -> float:
        """Returns the value of the rotation in radians."""
        return self.value

    def getDegrees(self) -> float:
        """Returns the value of the rotation in degrees."""
        return math.degrees(self.value)

    def tan(self) -> float:
        """Returns the tangent of the rotation."""
        return self.sin / self.cos



import math
from dataclasses import dataclass

from .rotation2d import Rotation2d


@dataclass(frozen=True)
class Translation2d:
    """Represents a translation in 2d space.

    This object can be used to represent a point or a vector.

    This assumes that you are using conventional mathematical axes.
    When the robot is placed on the origin, facing toward the X direction,
    moving forward increases the X, whereas moving to the left increases the Y.
    """

    #: The X component of the translation.
    x: float
    #: The Y component of the translation.
    y: float

    __slots__ = ("x", "y")

    def __init__(self, x: float = 0, y: float = 0):
        object.__setattr__(self, "x", x)
        object.__setattr__(self, "y", y)

    def getDistance(self, other: "Translation2d") -> float:
        """Calculates the distance between two translations in 2d space."""
        return math.hypot(other.x - self.x, other.y - self.y)

    def norm(self) -> float:
        """Returns the norm, or distance from the origin to the translation."""
        return math.hypot(self.x, self.y)

    def __abs__(self) -> float:
        """Returns the norm of this translation.

        This is equivalent to the norm method.
        """
        return self.norm()

    def rotateBy(self, other: Rotation2d) -> "Translation2d":
        """Applies a rotation to the translation in 2d space.

        This multiplies the translation vector by a counterclockwise
        rotation matrix of the given angle::

            [x_new]   [other.cos, -other.sin][x]
            [y_new] = [other.sin,  other.cos][y]

        For example, rotating a Translation2d(2, 0) by 90 degrees
        will return a Translation2d(0, 2).

        :param other: The rotation to rotate the translation by.

        :returns: The new rotated translation.
        """
        x = self.x
        y = self.y
        cos = other.cos
        sin = other.sin
        return Translation2d(x * cos - y * sin, x * sin + y * cos)

    def __add__(self, other: "Translation2d") -> "Translation2d":
        """Adds two translations in 2d space."""
        if not isinstance(other, Translation2d):
            return NotImplemented
        return Translation2d(self.x + other.x, self.y + other.y)

    def __sub__(self, other: "Translation2d") -> "Translation2d":
        """Subtracts the other translation from self."""
        if not isinstance(other, Translation2d):
            return NotImplemented
        return Translation2d(self.x - other.x, self.y - other.y)

    def __neg__(self) -> "Translation2d":
        """Takes the inverse of the translation."""
        return Translation2d(-self.x, -self.y)

    def __mul__(self, other: float) -> "Translation2d":
        """Multiplies the translation by a scalar."""
        if not isinstance(other, (float, int)):
            return NotImplemented
        return Translation2d(self.x * other, self.y * other)

    __rmul__ = __mul__

    def __truediv__(self, other: float) -> "Translation2d":
        """Divides the translation by a scalar."""
        if not isinstance(other, (float, int)):
            return NotImplemented
        return Translation2d(self.x / other, self.y / other)

    def __eq__(self, other: "Translation2d") -> bool:
        if not isinstance(other, Translation2d):
            return NotImplemented
        return math.isclose(self.x, other.x) and math.isclose(self.y, other.y)
